"""Wrapping dags"""

from meshed import DAG


class DDag(DAG):
    wrappers = ()

    def _call(self, *args, **kwargs):
        if not self.wrappers:
            return super()._call(*args, **kwargs)
        else:
            decorator = Line(*self.wrappers)
            decorated_dag_call = decorator(super()._call)
            return decorated_dag_call(*args, **kwargs)


def test_ddag():
    def f(a, b=2):
        return a + b

    def g(f, c=3):
        return f * c

    # d = DDag([f, g])
    d = DDag([f, g])

    d.dot_digraph()

    assert d(1, 2, 3) == 9  # can call
    from i2 import Sig

    assert str(Sig(d)) == '(a, b=2, c=3)'  # has correct signature

    def dec(func):
        def _dec(*args, **kwargs):
            print(func.__name__, args, kwargs)
            return func(*args, **kwargs)

        return _dec

    def rev(func):
        def _rev(*args, **kwargs):
            assert not kwargs, "Can't have keyword arguments with rev"
            return func(*args[::-1])

        return _rev

    d.wrappers = (dec,)

    assert d(1, 2, 3) == 9
    # prints: _call (1, 2, 3) {}
    assert d(1, 2, c=3) == 9
    # prints: _call (1, 2) {'c': 3}

    d.wrappers = (dec, rev)
    assert d(1, 2, 3) == 5
    # prints: _call (3, 2, 1) {}



"""
`meshed` contains a set of tools that allow the developer to provide a collection
of python objects (think functions) and some policy of how these should be connected
and get an aggregate object that will use the underlying objects in some way.

If you want something concrete, think of the python objects to be functions,
and the aggregation policies to be things like "function composition" (pipelines)
or DAGs.
But the intent is to be able to get more general aggregations than those.

Extras
------

`itools.py` contain tools that enable operations on graphs where graphs are represented
by an adjacency Mapping.

"""


from meshed.dag import DAG, ch_funcs, ch_names
from meshed.base import FuncNode
from meshed.makers import code_to_dag
from meshed.itools import random_graph, topological_sort
from meshed.util import (
    iterize,
    ConditionalIterize,
    instance_checker,
    replace_item_in_iterable,
)



"""
Tests for gk.py
"""

# Copyright 2016, Yahoo Inc.
# Licensed under the terms of the Apache License, Version 2.0. See the LICENSE file associated with the project for terms.
from contextlib import suppress

import math
import pickle

from pprint import pprint
from operator import add

with suppress(ModuleNotFoundError, ImportError):
    from numpy.testing import assert_raises

    from meshed.ext.gk import (
        operation,
        compose,
        Operation,
        optional,
        Network,
    )

    def test_network():

        # Sum operation, late-bind compute function
        sum_op1 = operation(name='sum_op1', needs=['a', 'b'], provides='sum_ab')(add)

        # sum_op1 is callable
        print(sum_op1(1, 2))

        # Multiply operation, decorate in-place
        @operation(name='mul_op1', needs=['sum_ab', 'b'], provides='sum_ab_times_b')
        def mul_op1(a, b):
            return a * b

        # mul_op1 is callable
        print(mul_op1(1, 2))

        # Pow operation
        @operation(
            name='pow_op1',
            needs='sum_ab',
            provides=['sum_ab_p1', 'sum_ab_p2', 'sum_ab_p3'],
            params={'exponent': 3},
        )
        def pow_op1(a, exponent=2):
            return [math.pow(a, y) for y in range(1, exponent + 1)]

        print(pow_op1._compute({'sum_ab': 2}, ['sum_ab_p2']))

        # Partial operation that is bound at a later time
        partial_op = operation(
            name='sum_op2', needs=['sum_ab_p1', 'sum_ab_p2'], provides='p1_plus_p2'
        )

        # Bind the partial operation
        sum_op2 = partial_op(add)

        # Sum operation, early-bind compute function
        sum_op_factory = operation(add)

        sum_op3 = sum_op_factory(name='sum_op3', needs=['a', 'b'], provides='sum_ab2')

        # sum_op3 is callable
        print(sum_op3(5, 6))

        # compose network
        net = compose(name='my network')(sum_op1, mul_op1, pow_op1, sum_op2, sum_op3)

        #
        # Running the network
        #

        # get all outputs
        pprint(net({'a': 1, 'b': 2}))

        # get specific outputs
        pprint(net({'a': 1, 'b': 2}, outputs=['sum_ab_times_b']))

        # start with inputs already computed
        pprint(net({'sum_ab': 1, 'b': 2}, outputs=['sum_ab_times_b']))

        # visualize network graph
        # net.plot(show=True)

    def test_network_simple_merge():

        sum_op1 = operation(name='sum_op1', needs=['a', 'b'], provides='sum1')(add)
        sum_op2 = operation(name='sum_op2', needs=['a', 'b'], provides='sum2')(add)
        sum_op3 = operation(name='sum_op3', needs=['sum1', 'c'], provides='sum3')(add)
        net1 = compose(name='my network 1')(sum_op1, sum_op2, sum_op3)
        pprint(net1({'a': 1, 'b': 2, 'c': 4}))

        sum_op4 = operation(name='sum_op1', needs=['d', 'e'], provides='a')(add)
        sum_op5 = operation(name='sum_op2', needs=['a', 'f'], provides='b')(add)
        net2 = compose(name='my network 2')(sum_op4, sum_op5)
        pprint(net2({'d': 1, 'e': 2, 'f': 4}))

        net3 = compose(name='merged')(net1, net2)
        pprint(net3({'c': 5, 'd': 1, 'e': 2, 'f': 4}))

    def test_network_deep_merge():

        sum_op1 = operation(name='sum_op1', needs=['a', 'b'], provides='sum1')(add)
        sum_op2 = operation(name='sum_op2', needs=['a', 'b'], provides='sum2')(add)
        sum_op3 = operation(name='sum_op3', needs=['sum1', 'c'], provides='sum3')(add)
        net1 = compose(name='my network 1')(sum_op1, sum_op2, sum_op3)
        pprint(net1({'a': 1, 'b': 2, 'c': 4}))

        sum_op4 = operation(name='sum_op1', needs=['a', 'b'], provides='sum1')(add)
        sum_op5 = operation(name='sum_op4', needs=['sum1', 'b'], provides='sum2')(add)
        net2 = compose(name='my network 2')(sum_op4, sum_op5)
        pprint(net2({'a': 1, 'b': 2}))

        net3 = compose(name='merged', merge=True)(net1, net2)
        pprint(net3({'a': 1, 'b': 2, 'c': 4}))

    def test_input_based_pruning():
        # Tests to make sure we don't need to pass graph inputs if we're provided
        # with data further downstream in the graph as an input.

        sum1 = 2
        sum2 = 5

        # Set up a net such that if sum1 and sum2 are provided directly, we don't
        # need to provide a and b.
        sum_op1 = operation(name='sum_op1', needs=['a', 'b'], provides='sum1')(add)
        sum_op2 = operation(name='sum_op2', needs=['a', 'b'], provides='sum2')(add)
        sum_op3 = operation(name='sum_op3', needs=['sum1', 'sum2'], provides='sum3')(
            add
        )
        net = compose(name='test_net')(sum_op1, sum_op2, sum_op3)

        results = net({'sum1': sum1, 'sum2': sum2})

        # Make sure we got expected result without having to pass a or b.
        assert 'sum3' in results
        assert results['sum3'] == add(sum1, sum2)

    def test_output_based_pruning():
        # Tests to make sure we don't need to pass graph inputs if they're not
        # needed to compute the requested outputs.

        c = 2
        d = 3

        # Set up a network such that we don't need to provide a or b if we only
        # request sum3 as output.
        sum_op1 = operation(name='sum_op1', needs=['a', 'b'], provides='sum1')(add)
        sum_op2 = operation(name='sum_op2', needs=['c', 'd'], provides='sum2')(add)
        sum_op3 = operation(name='sum_op3', needs=['c', 'sum2'], provides='sum3')(add)
        net = compose(name='test_net')(sum_op1, sum_op2, sum_op3)

        results = net({'c': c, 'd': d}, outputs=['sum3'])

        # Make sure we got expected result without having to pass a or b.
        assert 'sum3' in results
        assert results['sum3'] == add(c, add(c, d))

    def test_input_output_based_pruning():
        # Tests to make sure we don't need to pass graph inputs if they're not
        # needed to compute the requested outputs or of we're provided with
        # inputs that are further downstream in the graph.

        c = 2
        sum2 = 5

        # Set up a network such that we don't need to provide a or b d if we only
        # request sum3 as output and if we provide sum2.
        sum_op1 = operation(name='sum_op1', needs=['a', 'b'], provides='sum1')(add)
        sum_op2 = operation(name='sum_op2', needs=['c', 'd'], provides='sum2')(add)
        sum_op3 = operation(name='sum_op3', needs=['c', 'sum2'], provides='sum3')(add)
        net = compose(name='test_net')(sum_op1, sum_op2, sum_op3)

        results = net({'c': c, 'sum2': sum2}, outputs=['sum3'])

        # Make sure we got expected result without having to pass a, b, or d.
        assert 'sum3' in results
        assert results['sum3'] == add(c, sum2)

    def test_pruning_raises_for_bad_output():
        # Make sure we get a ValueError during the pruning step if we request an
        # output that doesn't exist.

        # Set up a network that doesn't have the output sum4, which we'll request
        # later.
        sum_op1 = operation(name='sum_op1', needs=['a', 'b'], provides='sum1')(add)
        sum_op2 = operation(name='sum_op2', needs=['c', 'd'], provides='sum2')(add)
        sum_op3 = operation(name='sum_op3', needs=['c', 'sum2'], provides='sum3')(add)
        net = compose(name='test_net')(sum_op1, sum_op2, sum_op3)

        # Request two outputs we can compute and one we can't compute.  Assert
        # that this raises a ValueError.
        assert_raises(
            ValueError,
            net,
            {'a': 1, 'b': 2, 'c': 3, 'd': 4},
            outputs=['sum1', 'sum3', 'sum4'],
        )

    def test_optional():
        # Test that optional() needs work as expected.

        # Function to add two values plus an optional third value.
        def addplusplus(a, b, c=0):
            return a + b + c

        sum_op = operation(
            name='sum_op1', needs=['a', 'b', optional('c')], provides='sum',
        )(addplusplus)

        net = compose(name='test_net')(sum_op)

        # Make sure output with optional arg is as expected.
        named_inputs = {'a': 4, 'b': 3, 'c': 2}
        results = net(named_inputs)
        assert 'sum' in results
        assert results['sum'] == sum(named_inputs.values())

        # Make sure output without optional arg is as expected.
        named_inputs = {'a': 4, 'b': 3}
        results = net(named_inputs)
        assert 'sum' in results
        assert results['sum'] == sum(named_inputs.values())

    def test_deleted_optional():
        # Test that DeleteInstructions included for optionals do not raise
        # exceptions when the corresponding input is not prodided.

        # Function to add two values plus an optional third value.
        def addplusplus(a, b, c=0):
            return a + b + c

        # Here, a DeleteInstruction will be inserted for the optional need 'c'.
        sum_op1 = operation(
            name='sum_op1', needs=['a', 'b', optional('c')], provides='sum1',
        )(addplusplus)
        sum_op2 = operation(name='sum_op2', needs=['sum1', 'sum1'], provides='sum2')(
            add
        )
        net = compose(name='test_net')(sum_op1, sum_op2)

        # DeleteInstructions are used only when a subset of outputs are requested.
        results = net({'a': 4, 'b': 3}, outputs=['sum2'])
        assert 'sum2' in results

    def test_parallel_execution():
        import time

        def fn(x):
            time.sleep(1)
            print('fn %s' % (time.time() - t0))
            return 1 + x

        def fn2(a, b):
            time.sleep(1)
            print('fn2 %s' % (time.time() - t0))
            return a + b

        def fn3(z, k=1):
            time.sleep(1)
            print('fn3 %s' % (time.time() - t0))
            return z + k

        pipeline = compose(name='l', merge=True)(
            # the following should execute in parallel under threaded execution mode
            operation(name='a', needs='x', provides='ao')(fn),
            operation(name='b', needs='x', provides='bo')(fn),
            # this should execute after a and b have finished
            operation(name='c', needs=['ao', 'bo'], provides='co')(fn2),
            operation(name='d', needs=['ao', optional('k')], provides='do')(fn3),
            operation(name='e', needs=['ao', 'bo'], provides='eo')(fn2),
            operation(name='f', needs='eo', provides='fo')(fn),
            operation(name='g', needs='fo', provides='go')(fn),
        )

        t0 = time.time()
        pipeline.set_execution_method('parallel')
        result_threaded = pipeline({'x': 10}, ['co', 'go', 'do'])
        print('threaded result')
        print(result_threaded)

        t0 = time.time()
        pipeline.set_execution_method('sequential')
        result_sequential = pipeline({'x': 10}, ['co', 'go', 'do'])
        print('sequential result')
        print(result_sequential)

        # make sure results are the same using either method
        assert result_sequential == result_threaded

    def test_multi_threading():
        import time
        import random
        from multiprocessing.dummy import Pool

        def op_a(a, b):
            time.sleep(random.random() * 0.02)
            return a + b

        def op_b(c, b):
            time.sleep(random.random() * 0.02)
            return c + b

        def op_c(a, b):
            time.sleep(random.random() * 0.02)
            return a * b

        pipeline = compose(name='pipeline', merge=True)(
            operation(name='op_a', needs=['a', 'b'], provides='c')(op_a),
            operation(name='op_b', needs=['c', 'b'], provides='d')(op_b),
            operation(name='op_c', needs=['a', 'b'], provides='e')(op_c),
        )

        def infer(i):
            # data = open("616039-bradpitt.jpg").read()
            outputs = ['c', 'd', 'e']
            results = pipeline({'a': 1, 'b': 2}, outputs)
            assert tuple(sorted(results.keys())) == tuple(sorted(outputs)), (
                outputs,
                results,
            )
            return results

        N = 100
        for i in range(20, 200):
            pool = Pool(i)
            pool.map(infer, range(N))
            pool.close()

    ####################################
    # Backwards compatibility
    ####################################

    # Classes must be defined as members of __main__ for pickleability

    # We first define some basic operations
    class Sum(Operation):
        def compute(self, inputs):
            a = inputs[0]
            b = inputs[1]
            return [a + b]

    class Mul(Operation):
        def compute(self, inputs):
            a = inputs[0]
            b = inputs[1]
            return [a * b]

    # This is an example of an operation that takes a parameter.
    # It also illustrates an operation that returns multiple outputs
    class Pow(Operation):
        def compute(self, inputs):

            a = inputs[0]
            outputs = []
            for y in range(1, self.params['exponent'] + 1):
                p = math.pow(a, y)
                outputs.append(p)
            return outputs

    def test_backwards_compatibility():

        sum_op1 = Sum(name='sum_op1', provides=['sum_ab'], needs=['a', 'b'])
        mul_op1 = Mul(
            name='mul_op1', provides=['sum_ab_times_b'], needs=['sum_ab', 'b']
        )
        pow_op1 = Pow(
            name='pow_op1',
            needs=['sum_ab'],
            provides=['sum_ab_p1', 'sum_ab_p2', 'sum_ab_p3'],
            params={'exponent': 3},
        )
        sum_op2 = Sum(
            name='sum_op2', provides=['p1_plus_p2'], needs=['sum_ab_p1', 'sum_ab_p2'],
        )

        net = Network()
        net.add_op(sum_op1)
        net.add_op(mul_op1)
        net.add_op(pow_op1)
        net.add_op(sum_op2)
        net.compile()

        # try the pickling part
        pickle.dumps(net)

        #
        # Running the network
        #

        # get all outputs
        pprint(net.compute(outputs=None, named_inputs={'a': 1, 'b': 2}))

        # get specific outputs
        pprint(net.compute(outputs=['sum_ab_times_b'], named_inputs={'a': 1, 'b': 2}))

        # start with inputs already computed
        pprint(
            net.compute(outputs=['sum_ab_times_b'], named_inputs={'sum_ab': 1, 'b': 2})
        )



from typing import Mapping
from collections import ChainMap
from meshed.itools import edge_reversed_graph
from i2 import Sig


def get_first_item_and_assert_unicity(seq):
    seq_length = len(seq)
    if seq_length:
        assert seq_length == 1, (
            f'There should be one and one only item in the ' f'sequence: {seq}'
        )
        return seq[0]
    else:
        return None


def func_node_names_and_outs(dag):
    for func_node in dag.func_nodes:
        yield func_node.name, func_node.out


NoSuchKey = type('NoSuchKey', (), {})

# TODO: Cache validation and invalidation
class CachedDag:
    def __init__(self, dag, cache=True, name=None):
        self.dag = dag
        self.reversed_graph = edge_reversed_graph(dag.graph_ids)
        self.roots = set(self.dag.roots)
        self.leafs = set(self.dag.leafs)
        self.var_nodes = set(self.dag.var_nodes)
        self.func_node_of_id = {fn.out: fn for fn in self.dag.func_nodes}
        self.name = name
        self.out_of_func_node_name = dict(func_node_names_and_outs(self.dag))
        self.defaults = Sig(self.dag).defaults
        if cache is True:
            self.cache = {}
        elif not isinstance(cache, Mapping):
            raise NotImplementedError(
                'This type of cache is not implemented (must resolve to a Mapping): '
                f'{cache=}'
            )
        self._cache = ChainMap(self.defaults, self.cache)

    @property
    def __name__(self):
        return self.name or self.dag.__name__

    def __iter__(self):
        yield from self.reversed_graph

    def func_node_id(self, k):
        func_node_name = get_first_item_and_assert_unicity(self.reversed_graph[k])
        if func_node_name is not None:
            return self.out_of_func_node_name[func_node_name]

    def __call__(self, k, input_kwargs=()):
        #         print(f"Calling ({k=},{input_kwargs=})\t{self.cache=}")
        if intersection := (input_kwargs.keys() & self.cache.keys()):
            raise ValueError(
                f"input_kwargs can't contain any keys that are already in cache! "
                f'These names were in both: {intersection}'
            )
        _cache = ChainMap(input_kwargs, self._cache)
        if k in _cache:
            return _cache[k]
        input_kwargs = dict(input_kwargs)
        func_node_id = self.func_node_id(k)
        #         print(f"{func_node_id=}")
        if func_node_id:
            if (output := self.cache.get(func_node_id)) is not None:
                return output
            else:
                func_node = self.func_node_of_id[func_node_id]
                input_sources = {
                    src: self(src, input_kwargs) for src in func_node.bind.values()
                }
                #                 inputs = dict(input_sources, **input_kwargs)  #
                #                 TODO: do we need to include **self.defaults in the
                #                  middle?
                inputs = ChainMap(_cache, input_sources)
                #                 print(f"Computing {func_node_id}: ", end=" ")
                output = func_node.call_on_scope(inputs, write_output_into_scope=False)
                self.cache[func_node_id] = output
                #                 print(f"result -> {output}")
                return output
        else:  # k is a root node
            inputs = ChainMap(input_kwargs, self._cache)
            assert k in self.roots, f'Was expecting this to be a root node: {k}'
            if (output := inputs.get(k, NoSuchKey)) is not NoSuchKey:
                return output
            else:
                raise TypeError(
                    f'The input_kwargs of a {self.__name__} call is missing 1 required '
                    f"argument: '{k}'"
                )


def cached_dag_test():
    def add(a, b=1):
        return a + b

    def mult(x, y=2):
        return x * y

    def exp(mult, n=3):
        return m ** n

    def subtract(a, b=4):
        return a - b

    from meshed import code_to_dag

    @code_to_dag(func_src=locals())
    def dag(w, ww, www):
        x = mult(w, ww)
        y = add(x, www)
        z = subtract(x, y)

    g = CachedDag(dag)

    assert g('z', {'w': 2, 'ww': 3, 'www': 4}) == -4 == dag(2, 3, 4)



import meshed as ms
import pytest


@pytest.fixture
def example_graph():
    return dict(a=['c'], b=['c', 'c'], c=['a', 'b', 'd', 'e'], d=['c'], e=['c', 'z'])


@pytest.fixture
def graph_children():
    return {0: [1, 2], 1: [2, 3, 4], 2: [1, 4], 3: [4]}


@pytest.fixture
def digraph_children():
    return {0: [1, 2], 1: [2, 3, 4], 2: [4], 3: [4]}


@pytest.fixture
def graph_dict():
    return dict(a='c', b='ce', c='abde', d='c', e=['c', 'z'], f={})


def test_add_edge(example_graph):
    g = example_graph
    ms.itools.add_edge(g, 'd', 'a')
    assert g == {
        'a': ['c'],
        'b': ['c', 'c'],
        'c': ['a', 'b', 'd', 'e'],
        'd': ['c', 'a'],
        'e': ['c', 'z'],
    }
    ms.itools.add_edge(g, 't', 'y')
    assert g == {
        'a': ['c'],
        'b': ['c', 'c'],
        'c': ['a', 'b', 'd', 'e'],
        'd': ['c', 'a'],
        'e': ['c', 'z'],
        't': ['y'],
    }


def test_edges(example_graph):
    g = example_graph
    assert sorted(ms.itools.edges(g)) == [
        ('a', 'c'),
        ('b', 'c'),
        ('b', 'c'),
        ('c', 'a'),
        ('c', 'b'),
        ('c', 'd'),
        ('c', 'e'),
        ('d', 'c'),
        ('e', 'c'),
        ('e', 'z'),
    ]


def test_nodes(example_graph):
    g = example_graph
    assert sorted(ms.itools.nodes(g)) == ['a', 'b', 'c', 'd', 'e', 'z']


def test_has_node():
    g = {0: [1, 2], 1: [2]}
    assert ms.itools.has_node(g, 0)
    assert ms.itools.has_node(g, 2)
    assert not ms.itools.has_node(g, 2, check_adjacencies=False)
    gg = {0: [1, 2], 1: [2], 2: []}
    assert ms.itools.has_node(gg, 2, check_adjacencies=False)


def test_successors(graph_children):
    g = graph_children
    assert set(ms.itools.successors(g, 1)) == {1, 2, 3, 4}
    assert set(ms.itools.successors(g, 3)) == {4}
    assert set(ms.itools.successors(g, 4)) == set()


def test_predecessors(graph_children):
    g = graph_children
    assert set(ms.itools.predecessors(g, 4)) == {0, 1, 2, 3}
    assert set(ms.itools.predecessors(g, 2)) == {0, 1, 2}
    assert set(ms.itools.predecessors(g, 0)) == set()


def test_children(graph_children):
    g = graph_children
    assert set(ms.itools.children(g, [2, 3])) == {1, 4}
    assert set(ms.itools.children(g, [4])) == set()


def test_parents(graph_children):
    g = graph_children
    assert set(ms.itools.parents(g, [2, 3])) == {0, 1}
    assert set(ms.itools.parents(g, [0])) == set()


def test_ancestors(digraph_children):
    g = digraph_children
    assert set(ms.itools.ancestors(g, [2, 3])) == {0, 1}
    assert set(ms.itools.ancestors(g, [0])) == set()


def test_descendants(digraph_children):
    g = digraph_children
    assert set(ms.itools.descendants(g, [2, 3])) == {4}
    assert set(ms.itools.descendants(g, [4])) == set()


def test_root_nodes(graph_dict):
    g = graph_dict
    assert sorted(ms.itools.root_nodes(g)) == ['f']


def test_leaf_nodes(graph_dict):
    g = graph_dict
    assert sorted(ms.itools.leaf_nodes(g)) == ['f', 'z']


def test_isolated_nodes(graph_dict):
    g = graph_dict
    assert set(ms.itools.isolated_nodes(g)) == {'f'}


def test_find_path(graph_dict):
    g = graph_dict
    assert ms.itools.find_path(g, 'a', 'c') == ['a', 'c']
    assert ms.itools.find_path(g, 'a', 'b') == ['a', 'c', 'b']
    assert ms.itools.find_path(g, 'a', 'z') == ['a', 'c', 'b', 'e', 'z']


def test_reverse_edges(example_graph):
    g = example_graph
    assert sorted(list(ms.itools.reverse_edges(g))) == [
        ('a', 'c'),
        ('b', 'c'),
        ('c', 'a'),
        ('c', 'b'),
        ('c', 'b'),
        ('c', 'd'),
        ('c', 'e'),
        ('d', 'c'),
        ('e', 'c'),
        ('z', 'e'),
    ]


def test_out_degrees(graph_dict):
    g = graph_dict
    assert dict(ms.itools.out_degrees(g)) == (
        {'a': 1, 'b': 2, 'c': 4, 'd': 1, 'e': 2, 'f': 0}
    )


def test_in_degrees(graph_dict):
    g = graph_dict
    assert dict(ms.itools.in_degrees(g)) == (
        {'a': 1, 'b': 1, 'c': 4, 'd': 1, 'e': 2, 'f': 0, 'z': 1}
    )


def test_copy_of_g_with_some_keys_removed(example_graph):
    g = example_graph
    keys = ['c', 'd']
    gg = ms.itools.copy_of_g_with_some_keys_removed(g, keys)
    assert gg == {'a': ['c'], 'b': ['c', 'c'], 'e': ['c', 'z']}


def test_topological_sort_helper(example_graph):
    g = example_graph
    v = 'a'
    stack = ['b', 'c']
    visited = set(['e'])
    ms.itools._topological_sort_helper(g, v, visited, stack)
    assert visited == {'a', 'b', 'c', 'd', 'e'}
    assert sorted(stack) == ['a', 'b', 'b', 'c', 'c', 'd']


def test_topological_sort():
    g = {0: [4, 2], 4: [3, 1], 2: [3], 3: [1]}
    assert list(ms.itools.topological_sort(g)) == [0, 2, 4, 3, 1]


def test_handle_exclude_nodes():
    def f(a=1, b=2, _exclude_nodes=['c']):
        return f'_exclude_nodes is now a set:{_exclude_nodes}'

    new_f = ms.itools._handle_exclude_nodes(f)
    assert new_f(1) == "_exclude_nodes is now a set:{'c'}"


@pytest.fixture
def simple_graph():
    return dict(a='c', b='cd', c='abd', e='')


def test_edge_reversed_graph(simple_graph):
    g = simple_graph
    assert ms.itools.edge_reversed_graph(g) == {
        'c': ['a', 'b'],
        'd': ['b', 'c'],
        'a': ['c'],
        'b': ['c'],
        'e': [],
    }
    reverse_g_with_sets = ms.itools.edge_reversed_graph(g, set, set.add)
    assert reverse_g_with_sets == {
        'c': {'a', 'b'},
        'd': {'b', 'c'},
        'a': {'c'},
        'b': {'c'},
        'e': set([]),
    }
    assert ms.itools.edge_reversed_graph(dict(e='', a='e')) == {
        'e': ['a'],
        'a': [],
    }
    assert ms.itools.edge_reversed_graph(dict(a='e', e='')) == {
        'e': ['a'],
        'a': [],
    }
